from . import attributes, construct, cleanup, defaults, load, recurse
from .. util import exception, json

ROOT_FILE = None


class Project:
    CHILDREN = 'maker', 'drivers', 'layout', 'animation'

    @staticmethod
    def pre_recursion(desc):
        return cleanup.cleanup(desc)

    def construct_child(self, datatype, typename=None, **kwds):
        construct = getattr(datatype, 'construct', None)
        if construct:
            return construct(self, **kwds)
        return datatype(**kwds)

    def __init__(self, *, drivers, layout, maker, path, animation, **kwds):
        attributes.check(kwds, 'project')
        self.path = path
        layout = layout or cleanup.cleanup_layout(animation)

        self.maker = self.construct_child(**maker)

        def post(desc):
            return self.construct_child(**desc)

        def create(root, name):
            with exception.add('Unable to create ' + name):
                return recurse.recurse(
                    root,
                    pre=None,
                    post=post,
                    python_path='bibliopixel.' + name)

        self.drivers = [create(d, 'drivers') for d in drivers]
        with exception.add('Unable to create layout'):
            self.layout = self.construct_child(**layout)
        self.animation = create(animation, 'animation')


def project(*descs, root_file=None):
    def default(o):
        if isinstance(o, type):
            return str(o)
        return str(o.__class__).replace('<class ', '<').replace('>', ' object>')

    desc = defaults.merge(*descs)

    global ROOT_FILE
    ROOT_FILE = root_file

    with load.extender(desc.get('path', '')):
        desc = recurse.recurse(desc)

    project = construct.construct(**desc)
    project.desc = desc
    return project



import os, sys, time, traceback

from . codes import CMDTYPE, LEDTYPE, SPIChipsets, BufferChipsets
from . devices import Devices
from .. channel_order import ChannelOrder
from .. driver_base import DriverBase
from ... util import log, util
from ... drivers.return_codes import (
    RETURN_CODES, print_error, raise_error, BiblioSerialError)


class Serial(DriverBase):
    """Main driver for Serial based LED strips"""

    def __init__(self, ledtype=None, num=0, dev="",
                 c_order=ChannelOrder.RGB, spi_speed=2,
                 gamma=None, restart_timeout=3,
                 device_id=None, hardwareID="1D50:60AB",
                 baudrate=921600, **kwds):

        if ledtype is None:
            raise ValueError('Must provide ledtype value!')

        if num == 0:
            raise ValueError('Must provide num value >0!')

        super().__init__(num, c_order=c_order, gamma=gamma, **kwds)
        self.devices = Devices(hardwareID, baudrate)
        self.serial = self.devices.serial

        if not (1 <= spi_speed <= 24 and ledtype in SPIChipsets):
            spi_speed = 1

        self._spi_speed = spi_speed
        self._com = None
        self._ledtype = ledtype
        self._bufPad = 0
        self.dev = dev
        self.device_version = 0
        self.device_id = device_id
        self._sync_packet = util.generate_header(CMDTYPE.SYNC, 0)

        if self.device_id is not None and not (0 <= self.device_id <= 255):
            raise ValueError("device_id must be between 0 and 255")

        resp = self._connect()
        if resp == RETURN_CODES.REBOOT:  # reboot needed
            log.info(
                "Reconfigure and reboot needed, waiting for controller to restart...")
            self._com.close()
            time.sleep(restart_timeout)
            resp = self._connect()
            if resp != RETURN_CODES.SUCCESS:
                raise_error(resp)
            else:
                log.info("Reconfigure success!")
        elif resp != RETURN_CODES.SUCCESS:
            raise_error(resp)

        if type in SPIChipsets:
            log.info("Using SPI Speed: %sMHz", self._spi_speed)

        self.set_device_brightness = self.set_brightness

    def cleanup(self):
        if self._com:
            log.info("Closing connection to: %s", self.dev)
            self._com.close()

    def _connect(self):
        try:
            self.devices.find_serial_devices()
            idv = self.devices.get_device(self.device_id)
            self.device_id, self.dev, self.device_version = idv
            try:
                self._com = self.serial.Serial(
                    self.dev, baudrate=self.devices.baudrate, timeout=5)
            except self.serial.SerialException as e:
                ports = self.devices.devices.values()
                error = "Invalid port specified. No COM ports available."
                if ports:
                    error = ("Invalid port specified. Try using one of: \n" +
                             "\n".join(ports))
                log.info(error)
                raise BiblioSerialError(error)

            packet = util.generate_header(CMDTYPE.SETUP_DATA, 4)
            packet.append(self._ledtype)  # set strip type
            byteCount = self.bufByteCount()
            if self._ledtype in BufferChipsets:
                if self._ledtype == LEDTYPE.APA102 and self.device_version >= 2:
                    pass
                else:
                    self._bufPad = BufferChipsets[
                        self._ledtype](self.numLEDs) * 3
                    byteCount += self._bufPad

            packet.append(byteCount & 0xFF)  # set 1st byte of byteCount
            packet.append(byteCount >> 8)  # set 2nd byte of byteCount
            packet.append(self._spi_speed)
            self._com.write(packet)

            resp = self._com.read(1)
            if len(resp) == 0:
                self.devices.error()

            return ord(resp)

        except self.serial.SerialException as e:
            error = ("Unable to connect to the device. Please check that "
                     "it is connected and the correct port is selected.")
            log.error(traceback.format_exc())
            log.error(error)
            raise e

    def set_brightness(self, brightness):
        super().set_brightness(brightness)
        packet = util.generate_header(CMDTYPE.BRIGHTNESS, 1)
        packet.append(self._brightness)
        self._com.write(packet)
        resp = ord(self._com.read(1))
        if resp == RETURN_CODES.SUCCESS:
            return True
        print_error(resp)

    def _send_packet(self):
        self._com.write(self._packet)

        resp = self._com.read(1)
        if len(resp) == 0:
            self.devices.error(fail=False)
        elif ord(resp) != RETURN_CODES.SUCCESS:
            print_error(ord(resp))
        else:
            self._com.flushInput()
            return True

    def _compute_packet(self):
        count = self.bufByteCount() + self._bufPad
        self._packet = util.generate_header(CMDTYPE.PIXEL_DATA, count)

        self._render()
        self._packet.extend(self._buf)
        self._packet.extend([0] * self._bufPad)

    def _send_sync(self):
        self._com.write(self._sync_packet)


# This is DEPRECATED.
DriverSerial = Serial


class TeensySmartMatrix(Serial):
    def __init__(self, width, height, dev="", device_id=None,
                 hardwareID="16C0:0483", **kwds):
        super().__init__(ledtype=LEDTYPE.GENERIC, num=width * height,
                         device_id=device_id, hardwareID=hardwareID, **kwds)
        self.sync = self._send_sync


# This is DEPRECATED.
DriverTeensySmartMatrix = TeensySmartMatrix



import argparse, os, sys
from . import common_flags
from .. util import log
from .. project.importer import import_module
from .. project import aliases, project


__all__ = ['main']

COMMANDS = (
    'all_pixel', 'all_pixel_test', 'clear_cache', 'color', 'demo',
    'devices', 'info', 'load', 'list', 'remove', 'reset', 'run', 'save', 'set',
    'show', 'update')

MODULES = {c: import_module('bibliopixel.main.' + c) for c in COMMANDS}


def no_command(*_):
    log.printer('ERROR: No command entered')
    log.printer('Valid:', ', '.join(COMMANDS))
    return -1


def get_args(argv=sys.argv):
    argv = ['-h' if a == '--help' else a for a in argv[1:]]

    # argparse doesn't give command-specific help for `bp --help <command>`
    # so we use `bp <command> --help` (#429)
    if len(argv) == 2 and argv[0] == '-h':
        argv.reverse()

    try:
        argv.remove('--version')
    except:
        pass
    else:
        log.printer('BiblioPixel version %s' % common_flags.VERSION)
        if not argv:
            return

    if argv and not argv[0].isidentifier() and '-h' not in argv:
        # The first argument can't be a command so try to run it.
        argv.insert(0, 'run')

    if argv and argv[0].startswith('-') and any(
            not a.startswith('-') for a in argv):
        log.printer(
            'bibliopixel: error: command line flags must appear at the end.')

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    for name, module in sorted(MODULES.items()):
        doc = module.__doc__
        subparser = subparsers.add_parser(name, help=doc)
        common_flags.add_common_flags(subparser)
        module.set_parser(subparser)
        subparser.description = doc + getattr(module, 'DESCRIPTION', '')

    if argv == ['-h']:
        log.printer(HELP)
    return parser.parse_args(argv)


def main():
    args = get_args()
    run = getattr(args, 'run', None)
    if not run:
        log.printer('ERROR: No command entered')
        log.printer('Valid commands are:')
        log.printer('    ', ', '.join(COMMANDS))
        log.printer()
        log.printer('For more help, type')
        log.printer()
        log.printer('    bp --help')
        sys.exit(-1)

    if args.verbose and args.loglevel != 'frame':
        log.set_log_level('debug')
    else:
        log.set_log_level(args.loglevel)

    try:
        return run(args) or 0
    except Exception as e:
        if args.verbose:
            raise
        log.printer('ERROR:', e.args[0], file=sys.stderr)
        log.printer(*e.args[1:], sep='\n', file=sys.stderr)
        result = getattr(e, 'errorcode', -1)

    sys.exit(result)


HELP = """\
bp - The BiblioPixel Project Runner.

`bp` is a command-line script installed with Bibliopixel.  It can run projects
and demos, configure hardware, save and load defaults, and more.

`bp` takes one of {0} *commands*:
  {1}
  {2}

Some commands take *arguments* - often names of files.

Examples of `bp` commands and arguments:

    bp info
    bp demo circle
    bp run foo.json bar.json
    bp foo.json bar.json      # run is the default command

The commands are summarized below, and you can get more detailed help about
each one by typing:

    bp <command> --help

or

    bp <command> -h

`bp` commands have *flags* that control how `bp` runs its commands.

Examples of flags are -v, --verbose, --defaults, -d, --simpixel.  Flags either
start with -, for one letter flags, or -- for multi-letter flags.

Sometimes two flags have the same meaning, like --simpixel and -s,
--verbose and -v, or --dimensions and --dim.

Some flags take an argument, like --dim=160 or --loglevel=frame.
Other flags do not, like --verbose or -v.

A `bp` command line can optionally include a command, a list of arguments,
and a list of flags:

    bp [<command>] [<argument> ...] [<flag> ...]

Examples of `bp` command lines:

    bp info --help
    bp run -h
    bp info
    bp color "indian red 4"
    bp foo.json bar.json -v
    bp foo.json bar.json -v --numbers=float --dump --defaults=default.json
""".format(len(COMMANDS),
           ', '.join(COMMANDS[0:8]),
           ', '.join(COMMANDS[8:]))



import time
from .. import util
from .. project import attributes, data_maker, fields
from .. util.threads.update_threading import UpdateThreading


class Layout(object):
    pre_recursion = fields.default_converter

    @classmethod
    def construct(cls, project, **desc):
        """Construct a layout."""
        return cls(project.drivers, maker=project.maker, **desc)

    def __init__(self, drivers, threadedUpdate, brightness,
                 maker=data_maker.MAKER, **kwds):
        """Base LED class. Use Strip or Matrix instead!"""
        attributes.set_reserved(self, 'layout', **kwds)
        self.drivers = drivers if isinstance(drivers, list) else [drivers]
        self.maker = maker

        if not hasattr(self, 'numLEDs'):
            self.numLEDs = sum(d.numLEDs for d in self.drivers)

        # This buffer will always be the same list - i.e. is guaranteed to only
        # be changed by list surgery, never assignment.
        self._colors = maker.color_list(self.numLEDs)

        pos = 0
        for d in self.drivers:
            d.set_colors(self._colors, pos)
            pos += d.numLEDs

        self.frame_render_time = 0
        self.animation_sleep_time = None

        self.threading = UpdateThreading(threadedUpdate, self)
        self.brightness = brightness
        self.set_brightness(brightness)

    def set_pixel_positions(self, pixel_positions):
        for d in self.drivers:
            d.set_pixel_positions(pixel_positions)

    def update(self):
        """DEPRECATED - use self.push_to_driver()"""
        return self.push_to_driver()

    def start(self):
        for d in self.drivers:
            d.start()

    def cleanup_drivers(self):
        for d in self.drivers:
            d.cleanup()

    def cleanup(self):
        self.all_off()
        self.push_to_driver()
        self.threading.wait_for_update()

    @property
    def color_list(self):
        return self._colors

    @color_list.setter
    def color_list(self, cl):
        self._colors[:] = cl

    def _get_base(self, pixel):
        if pixel >= 0 and pixel < self.numLEDs:
            return self._colors[pixel]
        return 0, 0, 0  # don't go out of bounds

    def _set_base(self, pixel, color):
        if pixel >= 0 and pixel < self.numLEDs:
            if isinstance(color, str):
                color = util.colors.COLORS[color]
            else:
                color = tuple(color)
            self.color_list[pixel] = color

    def get_pixel_positions(self):
        result = []
        for x in range(len(self.numLEDs)):
            result.append([x, 0, 0])
        return result

    def push_to_driver(self):
        """Push the current pixel state to the driver"""
        # This is overridden elsewhere.
        self.threading.push_to_driver()

    # use with caution!
    def set_colors(self, buf):
        """Use with extreme caution!
        Directly sets the internal buffer and bypasses all brightness and rotation control.
        buf must also be in the exact format required by the display type.
        """
        if len(self._colors) != len(buf):
            raise IOError("Data buffer size incorrect! "
                          "Expected: {} bytes / Received: {} bytes"
                          .format(len(self._colors), len(buf)))
        self._colors[:] = buf

    def setBuffer(self, buf):
        """DEPRECATED!"""
        # https://stackoverflow.com/questions/1624883
        self.set_colors(buf=list(zip(*(iter(buf),) * 3)))

    def set_brightness(self, brightness):
        self.brightness = brightness
        for d in self.drivers:
            d.set_brightness(brightness)

    # Set single pixel to RGB value
    def setRGB(self, pixel, r, g, b):
        """Set single pixel using individual RGB values instead of tuple"""
        self._set_base(pixel, (r, g, b))

    def setHSV(self, pixel, hsv):
        """Set single pixel to HSV tuple"""
        color = util.colors.hsv2rgb(hsv)
        self._set_base(pixel, color)

    # turns off the desired pixel
    def setOff(self, pixel):
        """Set single pixel off"""
        self._set_base(pixel, (0, 0, 0))

    def all_off(self):
        """Set all pixels off"""
        self._colors[:] = [(0, 0, 0)] * self.numLEDs

    # Fill the strand (or a subset) with a single color using a Color object
    def fill(self, color, start=0, end=-1):
        """Fill the entire strip with RGB color tuple"""
        start = max(start, 0)
        if end < 0 or end >= self.numLEDs:
            end = self.numLEDs - 1
        for led in range(start, end + 1):  # since 0-index include end in range
            self._set_base(led, color)

    # Fill the strand (or a subset) with a single color using RGB values
    def fillRGB(self, r, g, b, start=0, end=-1):
        """Fill entire strip by giving individual RGB values instead of tuple"""
        self.fill((r, g, b), start, end)

    # Fill the strand (or a subset) with a single color using HSV values
    def fillHSV(self, hsv, start=0, end=-1):
        """Fill the entire strip with HSV color tuple"""
        self.fill(util.colors.hsv2rgb(hsv), start, end)


class MultiLayout(Layout):
    def __init__(self, *args, gen_coord_map=None, coord_map=None, **kwds):
        super().__init__(*args, **kwds)
        if gen_coord_map:
            if coord_map:
                util.log.warning('Cannot set both coord_map and gen_coord_map')
            elif isinstance(gen_coord_map, dict):
                coord_map = self.gen_multi(**gen_coord_map)
            else:
                coord_map = self.gen_multi(gen_coord_map)
        self.coord_map = coord_map

    def gen_multi(self, *args, **kwds):
        raise NotImplementedError
